#!/usr/bin/env python3
"""Stuff related to vectors/paths within a gimp document.
"""
# from .GimpIOBase import GimpIOBase
from __future__ import annotations

from binaryiotools import IO

from .GimpParasites import GimpParasite


class GimpVector:
	"""A gimp brush stroke vector."""

	def __init__(self, parent):
		# GimpIOBase.__init__(self, parent)
		_ = parent
		self.name = ""
		self.uniqueId = 0
		self.visible = True
		self.linked = False
		self.parasites = []
		self.strokes = []

	def decode(self, data: bytes, index: int = 0) -> int:
		"""Decode a byte buffer.

		Args:
			data (bytes): data buffer to decode
			index (int, optional): index within the buffer to start at. Defaults to 0.

		Returns:
			int: offset
		"""
		ioBuf = IO(data, index, boolSize=32)
		self.name = ioBuf.sz754
		self.uniqueId = ioBuf.u32
		self.visible = ioBuf.boolean
		self.linked = ioBuf.boolean
		numParasites = ioBuf.u32
		numStrokes = ioBuf.u32
		for _ in range(numParasites):
			parasite = GimpParasite()
			ioBuf.index = parasite.decode(ioBuf.data, ioBuf.index)
			self.parasites.append(parasite)
		for _ in range(numStrokes):
			gimpstroke = GimpStroke(self)
			ioBuf.index = gimpstroke.decode(ioBuf.data, ioBuf.index)
			self.strokes.append(gimpstroke)
		return ioBuf.index

	def encode(self):
		"""Encode to binary data."""
		ioBuf = IO(boolSize=32)
		ioBuf.sz754 = self.name
		ioBuf.u32 = self.uniqueId
		ioBuf.boolean = self.visible
		ioBuf.boolean = self.linked
		ioBuf.u32 = len(self.parasites)
		ioBuf.u32 = len(self.strokes)
		for parasite in self.parasites:
			ioBuf.addBytes(parasite.encode())
		for gimpstroke in self.strokes:
			ioBuf.addBytes(gimpstroke.encode())
		return ioBuf.data

	def __repr__(self, indent: str = "") -> str:
		"""Get a textual representation of this object."""
		ret = []
		ret.append(f"Name: {self.name}")
		ret.append(f"Unique ID (tattoo): {self.uniqueId}")
		ret.append(f"Visible: {self.visible}")
		ret.append(f"Linked: {self.linked}")
		if self.parasites:
			ret.append("Parasites: ")
			for item in self.parasites:
				ret.append(item.__repr__(indent + "\t"))
		if self.strokes:
			ret.append("Strokes: ")
			for item in self.strokes:
				ret.append(item.__repr__(indent + "\t"))
		return indent + ((f"\n{indent}").join(ret))


class GimpStroke:
	"""A single stroke within a vector."""

	STROKE_TYPES = ["None", "Bezier"]

	def __init__(self, parent):
		# GimpIOBase.__init__(self, parent)
		_ = parent
		self.strokeType = 1  # one of self.STROKE_TYPES
		self.closedShape = True
		self.points = []

	def decode(self, data: bytes, index: int = 0) -> int:
		"""Decode a byte buffer.

		Args:
			data (bytes): data buffer to decode
			index (int, optional): index within the buffer to start at. Defaults to 0.

		Returns:
			int: offset
		"""
		ioBuf = IO(data, index, boolSize=32)
		self.strokeType = ioBuf.u32
		self.closedShape = ioBuf.boolean
		numFloatsPerPoint = ioBuf.u32
		numPoints = ioBuf.u32
		for _ in range(numPoints):
			gimpPoint = GimpPoint(self)
			ioBuf.index = gimpPoint.decode(ioBuf.data, ioBuf.index, numFloatsPerPoint)
			self.points.append(gimpPoint)
		return ioBuf.index

	def encode(self):
		"""Encode to binary data."""
		ioBuf = IO(boolSize=32)
		ioBuf.u32 = self.strokeType
		ioBuf.boolean = self.closedShape
		# ioBuf.u32 = numFloatsPerPoint
		# ioBuf.u32 = numPoints
		for gimpPoint in self.points:
			ioBuf.addBytes(gimpPoint.encode())
		return ioBuf.data

	def __repr__(self, indent: str = ""):
		"""Get a textual representation of this object."""
		ret = []
		ret.append(f"Stroke Type: {self.STROKE_TYPES[self.strokeType]}")
		ret.append(f"Closed: {self.closedShape}")
		ret.append("Points: ")
		for point in self.points:
			ret.append(point.__repr__(indent + "\t"))
		return indent + ((f"\n{indent}").join(ret))


class GimpPoint:
	"""A single point within a stroke."""

	POINT_TYPES = ["Anchor", "Bezier control point"]

	def __init__(self, parent):
		_ = parent
		self.x = 0
		self.y = 0
		self.pressure = 1.0
		self.xTilt = 0.5
		self.yTilt = 0.5
		self.wheel = 0.5
		self.pointType = 0

	def decode(self, data: bytes, index: int = 0, numFloatsPerPoint: int = 0):
		"""Decode a byte buffer.

		Args:
			data (bytes): data buffer to decode
			index (int, optional): index within the buffer to start at. Defaults to 0.
			numFloatsPerPoint (int, optional): required so we know
			how many different brush dynamic measurements are
			inside each point. Defaults to 0.

		Returns:
			int: offset
		"""
		ioBuf = IO(data, index, boolSize=32)
		self.pressure = 1.0
		self.xTilt = 0.5
		self.yTilt = 0.5
		self.wheel = 0.5
		self.pointType = ioBuf.u32
		if numFloatsPerPoint < 1:
			numFloatsPerPoint = int((len(ioBuf.data) - ioBuf.index) / 4)
		self.x = ioBuf.float32
		self.y = ioBuf.float32
		if numFloatsPerPoint > 2:
			self.pressure = ioBuf.float32
			if numFloatsPerPoint > 3:
				self.xTilt = ioBuf.float32
				if numFloatsPerPoint > 4:
					self.yTilt = ioBuf.float32
					if numFloatsPerPoint > 5:
						self.wheel = ioBuf.float32
		return ioBuf.index

	def encode(self):
		"""Encode to binary data."""
		ioBuf = IO(boolSize=32)
		ioBuf.u32 = self.pointType
		ioBuf.float32 = self.x
		ioBuf.float32 = self.y
		if self.pressure is not None:
			ioBuf.float32 = self.pressure
			if self.xTilt is not None:
				ioBuf.float32 = self.xTilt
				if self.yTilt is not None:
					ioBuf.float32 = self.yTilt
					if self.wheel is not None:
						ioBuf.float32 = self.wheel
		return ioBuf.data

	def __repr__(self, indent=""):
		"""Get a textual representation of this object."""
		ret = []
		ret.append(f"Location: ({self.x}" + f",{self.y})")
		ret.append(f"Pressure: {self.pressure}")
		ret.append(f"Location: ({self.xTilt}" + f",{self.yTilt})")
		ret.append(f"Wheel: {self.wheel}")
		return indent + ((f"\n{indent}").join(ret))



#!/usr/bin/env python3
"""Pure python implementation of the gimp gpl palette format.
"""
from __future__ import annotations

from io import BytesIO

from . import utils


class GimpGplPalette:
	"""Pure python implementation of the gimp gpl palette format."""

	def __init__(self, fileName: BytesIO | str | None = None):
		"""Pure python implementation of the gimp gpl palette format.

		Args:
			fileName (BytesIO, str, optional): filename. Defaults to None.
		"""
		self.name = ""
		self.columns = 16
		self.colors = []
		self.colorNames = []
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: BytesIO | str):
		"""Load a gimp file.

		:param fileName: can be a file name or a file-like object
		"""
		self.fileName, data = utils.fileOpen(fileName)
		self.decode(data.decode("utf-8"))

	def decode(self, data: str) -> None:
		"""Decode a byte buffer.

		Args:
			data (str): data buffer to decode

		Raises:
			RuntimeError: File format error.  Magic value mismatch.
		"""
		lines = [s.strip() for s in data.split("\n")]
		if lines[0] != "GIMP Palette":
			raise RuntimeError("File format error.  Magic value mismatch.")
		self.name = lines[1].split(":", 1)[-1].lstrip()
		self.columns = int(lines[2].split(":", 1)[-1].lstrip())
		for line in lines[3:]:
			if len(line) < 1 or line[0] == "#":  # Commented Line
				continue
			line = line.split(None, 4)
			if len(line) < 3:
				continue
			self.colors.append((int(line[0]), int(line[1]), int(line[2])))
			if len(line) > 3:
				self.colorNames.append(line[3])
			else:
				self.colorNames.append(None)

	def encode(self):
		"""Encode to a raw data stream."""
		data = []
		data.append("GIMP Palette")
		data.append(f"Name: {self.name}")
		data.append(f"Columns: {self.columns}")
		data.append("#")
		for i, color in enumerate(self.colors):
			colorName = self.colorNames[i]
			line = (
				str(color[0]).rjust(3) + " " + str(color[1]).rjust(3) + " " + str(color[2]).rjust(3)
			)
			if colorName is not None:
				line = line + "\t" + colorName
			data.append(line)
		return ("\n".join(data) + "\n").encode("utf-8")

	def save(self, fileName: str | BytesIO):
		"""Save this gimp image to a file."""
		utils.save(self.encode(), fileName)

	def __repr__(self):
		"""Get a textual representation of this object."""
		ret = []
		if self.fileName is not None:
			ret.append(f"fileName: {self.fileName}")
		ret.append(f"Name: {self.name}")
		ret.append(f"Columns: {self.columns}")
		ret.append("Colors:")
		for i, color in enumerate(self.colors):
			colorName = self.colorNames[i]
			line = f"{color[0]},{color[1]},{color[2]}"
			if colorName is not None:
				line = line + " " + colorName
		return "\n".join(ret)

	def __eq__(self, other: GimpGplPalette):
		"""Perform a comparison."""
		if other.name != self.name:
			return False
		if other.columns != self.columns:
			return False
		if len(self.colors) != len(other.colors):
			return False
		for i, colour in enumerate(self.colors):
			if colour != other.colors[i]:
				return False
			if self.colorNames[i] != other.colorNames[i]:
				return False
		return True



#!/usr/bin/env python3
"""Gimp color gradient.
"""
from __future__ import annotations

from io import BytesIO

from . import utils


class GradientSegment:
	"""Single segment within a gradient."""

	BLEND_FUNCTIONS = [
		"linear",
		"curved",
		"sinusoidal",
		"spherical (increasing)",
		"spherical (decreasing)",
		"step",
	]
	COLOR_TYPES = ["RGB", "HSV CCW", "HSV CW"]
	ENDPOINT_COLOR_TYPES = [
		"fixed",
		"foreground",
		"foreground transparent",
		"background",
		"background transparent",
	]

	def __init__(self):
		"""Single segment within a gradient."""
		self.leftPosition = 0
		self.middlePosition = 0.5
		self.rightPosition = 1.0
		self.leftColor = (0, 0, 0, 0)
		self.rightColor = (255, 255, 255, 0)
		self.blendFunc = None  # one of self.BLEND_FUNCTIONS
		self.colorType = None  # one of self.COLOR_TYPES
		self.leftColorType = None  # one of self.ENDPOINT_COLOR_TYPES
		self.rightColorType = None  # one of self.ENDPOINT_COLOR_TYPES

	def getColor(self, percent: float):
		"""Given a decimal percent (1.0 = 100%) retrieve the appropriate color
		for this point in the gradient.
		"""
		_ = self, percent
		raise NotImplementedError()

	def decode(self, dataIn: str):
		"""Decode a byte buffer.

		Args:
			dataIn (str): data buffer to decode

		Raises:
			RuntimeError: [description]
		"""
		data = dataIn.split(" ")
		if len(data) < 11 or len(data) > 15:
			raise RuntimeError(f"Data table is unexpected size. {len(data)}")
		self.leftPosition = float(data[0])
		self.middlePosition = float(data[1])
		self.rightPosition = float(data[2])
		self.leftColor = (float(data[3]), float(data[4]), float(data[5]), float(data[6]))
		self.rightColor = (float(data[7]), float(data[8]), float(data[9]), float(data[10]))
		if len(data) >= 12:
			self.blendFunc = int(data[11])
			if len(data) >= 13:
				self.colorType = int(data[12])
				if len(data) >= 14:
					self.leftColorType = int(data[13])
					if len(data) >= 15:
						self.rightColorType = int(data[14])

	def encode(self):
		"""Encode this to a byte array."""
		ret = []
		ret.append(f"{self.leftPosition:06f}")
		ret.append(f"{self.middlePosition:06f}")
		ret.append(f"{self.rightPosition:06f}")
		for chan in self.leftColor:
			ret.append(f"{chan:06f}")
		for chan in self.rightColor:
			ret.append(f"{chan:06f}")
		if self.blendFunc is not None:
			ret.append(f"{self.blendFunc}")
			if self.colorType is not None:
				ret.append(f"{self.colorType}")
				if self.leftColorType is not None:
					ret.append(f"{self.leftColorType}")
					if self.rightColorType is not None:
						ret.append(f"{self.rightColorType}")
		return " ".join(ret)

	def __repr__(self, indent=""):
		"""Get a textual representation of this object."""
		ret = []
		ret.append(f"Left Position: {self.leftPosition}")
		ret.append(f"Middle Position: {self.middlePosition}")
		ret.append(f"Right Position: {self.rightPosition}")
		ret.append(f"Left Color: {self.leftColor}")
		ret.append(f"Right Color: {self.rightColor}")
		if self.blendFunc:
			ret.append(f"Blend Function: {self.BLEND_FUNCTIONS[self.blendFunc]}")
		if self.colorType:
			ret.append(f"Color Type: {self.COLOR_TYPES[self.colorType]}")
		if self.leftColorType:
			ret.append(f"Left Color Type: {self.ENDPOINT_COLOR_TYPES[self.leftColorType]}")
		if self.rightColorType:
			ret.append(f"Right Color Type: {self.ENDPOINT_COLOR_TYPES[self.rightColorType]}")
		return (f"\n{indent}").join(ret)


class GimpGgrGradient:
	"""Gimp color gradient.

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/ggr.txt
	"""

	def __init__(self, fileName: str | None = None):
		"""GimpGgrGradient.

		Args:
			fileName (str, optional): filename. Defaults to None.
		"""
		self.fileName = None
		self.segments = []
		self.name = ""
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: BytesIO | str):
		"""Load a gimp file.

		:param fileName: can be a file name or a file-like object
		"""
		self.fileName, data = utils.fileOpen(fileName)
		self.decode(data)

	def decode(self, dataIn: bytes):
		"""Decode a byte buffer.

		Args:
			dataIn (bytes): data buffer to decode

		Raises:
			RuntimeError: File format error.  Magic value mismatch.
		"""
		data = dataIn.decode("utf-8").split("\n")
		data = [l.strip() for l in data]
		if data[0] != "GIMP Gradient":
			raise RuntimeError("File format error.  Magic value mismatch.")
		self.name = data[1].split(":", 1)[-1].strip()
		numSegments = int(data[2])
		for i in range(numSegments):
			gSeg = GradientSegment()
			gSeg.decode(data[i + 3])
			self.segments.append(gSeg)

	def encode(self):
		"""Encode this to a byte array."""
		ret = ["GIMP Gradient"]
		ret.append(f"Name: {self.name}")
		ret.append(str(len(self.segments)))
		for segment in self.segments:
			ret.append(segment.encode())
		return ("\n".join(ret) + "\n").encode("utf-8")

	def save(self, tofileName=None):
		"""Save this gimp image to a file."""
		utils.save(self.encode(), tofileName)

	def getColor(self, percent):
		"""Given a decimal percent (1.0 = 100%) retrieve...

		the appropriate color for this point in the gradient.
		"""
		raise NotImplementedError()

	def __repr__(self, indent=""):
		"""Get a textual representation of this object."""
		ret = []
		if self.fileName is not None:
			ret.append(f"fileName: {self.fileName}")
		ret.append(f"Name: {self.name}")
		for seg in self.segments:
			ret.append(seg.__repr__(indent + "\t"))
		return (f"\n{indent}").join(ret)



"""Represents a single layer in a gimp image.
"""
from __future__ import annotations

from binaryiotools import IO
from PIL.Image import Image

from .GimpChannel import GimpChannel
from .GimpImageHierarchy import GimpImageHierarchy
from .GimpIOBase import GimpIOBase


class GimpLayer(GimpIOBase):
	"""Represents a single layer in a gimp image."""

	COLOR_MODES = [
		"RGB color without alpha",
		"RGB color with alpha",
		"Grayscale without alpha",
		"Grayscale with alpha",
		"Indexed without alpha",
		"Indexed with alpha",
	]
	PIL_MODE_TO_LAYER_MODE = {"L": 2, "LA": 3, "RGB": 0, "RGBA": 1}

	def __init__(self, parent, name: str | None = None, image: Image | None = None):
		"""Represents a single layer in a gimp image.

		Args:
			parent ([type]): some parent node/ layer
			name (str, optional): name of the layer. Defaults to None.
			image (Image, optional): image stored in the layer. Defaults to None.
		"""
		GimpIOBase.__init__(self, parent)
		self.width = 0
		self.height = 0
		self.colorMode = 0
		self.name = name
		self._imageHierarchy = None
		self._imageHierarchyPtr = None
		self._mask = None
		self._maskPtr = None
		self._data = None
		if image is not None:
			self.image = image  # done last as it resets some of the above defaults

	def decode(self, data: bytes, index: int = 0) -> int:
		"""Decode a byte buffer.

		Steps:
		Create a new IO buffer (array of binary values)
		Grab attributes as outlined in the spec
		List of properties
		Get the image hierarchy and mask pointers
		Return the offset

		Args:
			data (bytes): data buffer to decode
			index (int, optional): index within the buffer to start at]. Defaults to 0.

		Returns:
			int: offset
		"""
		# Create a new IO buffer (array of binary values)
		ioBuf = IO(data, index)
		# Grab attributes as outlined in the spec
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		self.colorMode = ioBuf.u32  # one of self.COLOR_MODES
		self.name = ioBuf.sz754
		# List of properties
		self._propertiesDecode(ioBuf)
		# Get the image hierarchy and mask pointers
		self._imageHierarchyPtr = self._pointerDecode(ioBuf)
		self._maskPtr = self._pointerDecode(ioBuf)
		self._mask = None
		self._data = data
		# Return the offset
		return ioBuf.index

	def encode(self):
		"""Encode to byte array.

		Steps:
		Create a new IO buffer (array of binary values)
		Set attributes as outlined in the spec
		List of properties
		Set the image hierarchy and mask pointers
		Return the data

		"""
		# Create a new IO buffer (array of binary values)
		dataAreaIO = IO()
		ioBuf = IO()
		# Set attributes as outlined in the spec
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		ioBuf.u32 = self.colorMode
		ioBuf.sz754 = self.name
		# Layer properties
		ioBuf.addBytes(self._propertiesEncode())
		# Pointer to the image heirachy structure
		dataAreaIndex = ioBuf.index + self.pointerSize * 2
		ioBuf.addBytes(self._pointerEncode(dataAreaIndex))
		dataAreaIO.addBytes(self.imageHierarchy.encode())
		# ioBuf.addBytes(self._pointerEncode_(dataAreaIndex))
		# Pointer to the layer mask
		if self.mask is not None:
			dataAreaIO.addBytes(self.mask.encode())
		ioBuf.addBytes(self._pointerEncode(dataAreaIndex + dataAreaIO.index))
		ioBuf.addBytes(dataAreaIO)
		# Return the data
		return ioBuf.data

	@property
	def mask(self):
		"""Get the layer mask."""
		if self._mask is None and self._maskPtr is not None and self._maskPtr != 0:
			self._mask = GimpChannel(self)
			if self._data:
				self._mask.decode(self._data, self._maskPtr)
		return self._mask

	@property
	def image(self) -> Image | None:
		"""Get the layer image.

		NOTE: can return None!
		"""
		if self.imageHierarchy is None:
			return None
		return self.imageHierarchy.image

	@image.setter
	def image(self, image: Image):
		"""Set the layer image.

		NOTE: resets layer width, height, and colorMode
		"""
		self.height = image.height
		self.width = image.width
		if image.mode not in self.PIL_MODE_TO_LAYER_MODE:
			raise NotImplementedError('No way of handlng PIL image mode "' + image.mode + '"')
		self.colorMode = self.PIL_MODE_TO_LAYER_MODE[image.mode]
		if not self.name and isinstance(image, str):
			# try to use a fileName as the name
			self.name = image.rsplit("\\", 1)[-1].rsplit("/", 1)[-1]
		self._imageHierarchy = GimpImageHierarchy(self)
		self._imageHierarchy.image = image

	@property
	def imageHierarchy(self) -> GimpImageHierarchy:
		"""Get the image hierarchy objects.

		This is mainly needed for deciphering image, and therefore,
		of little use to you, the user.

		NOTE: can return None if it has been fully read into an image
		"""
		if self._imageHierarchy is None and self._imageHierarchyPtr:
			self._imageHierarchy = GimpImageHierarchy(self)
		if self._imageHierarchy and self._data and self._imageHierarchyPtr:
			self._imageHierarchy.decode(self._data, self._imageHierarchyPtr)
			return self._imageHierarchy
		raise RuntimeError("self._imageHierarchy or self._data or self._imageHierarchyPtr is None")

	@imageHierarchy.setter
	def imageHierarchy(self, imgHierarchy):
		"""Set the image hierarchy."""
		self._imageHierarchy = imgHierarchy

	def forceFullyLoaded(self):
		"""Make sure everything is fully loaded from the file."""
		if self.mask is not None:
			self.mask.forceFullyLoaded()
		_ = self.image  # make sure the image is loaded so we can delete the hierarchy nonsense
		# self._imageHierarchy = None
		# self._data = None

	def __repr__(self, indent=""):
		"""Get a textual representation of this object."""
		ret = []
		ret.append(f"Name: {self.name}")
		ret.append(f"Size: {self.width} x {self.height}")
		ret.append("colorMode: " + self.COLOR_MODES[self.colorMode])
		ret.append(GimpIOBase.__repr__(self, indent))
		mask = self.mask
		if mask is not None:
			ret.append("Mask:")
			ret.append(mask.__repr__(indent + "\t"))
		return indent + ((f"\n{indent}").join(ret))



"""A specialized binary file base for Gimp files.
"""
from __future__ import annotations

import struct

from binaryiotools import IO

from .GimpParasites import GimpParasite
from .GimpVectors import GimpVector


class GimpIOBase:
	"""A specialized binary file base for Gimp files."""

	COLOR_MODES = ["RGB", "Grayscale", "Indexed"]
	UNITS = ["Inches", "Millimeters", "Points", "Picas"]
	UNITS_TO_MM = [25.4, 1, 127 / 360.0, 127 / 30.0]
	COMPOSITE_MODES = ["Union", "Clip to backdrop", "Clip to layer", "Intersection"]
	COMPOSITE_SPACES = ["RGB (linear)", "RGB (perceptual)", "LAB"]
	TAG_COLORS = ["None", "Blue", "Green", "Yellow", "Orange", "Brown", "Red", "Violet", "Gray"]
	COMPRESSION_MODES = ["None", "RLE", "Zlib", "Fractal"]

	BLEND_MODES = [
		"Normal (legacy)",
		"Dissolve (legacy)",
		"Behind (legacy)",
		"Multiply (legacy)",
		"Screen (legacy)",
		"Old broken Overlay",
		"Difference (legacy)",
		"Addition (legacy)",
		"Subtract (legacy)",
		"Darken only (legacy)",
		"Lighten only (legacy)",
		"Hue (HSV) (legacy)",
		"Saturation (HSV) (legacy)",
		"Color (HSL) (legacy)",
		"Value (HSV) (legacy)",
		"Divide (legacy)",
		"Dodge (legacy)",
		"Burn (legacy)",
		"Hard Light (legacy)",
		"Soft light (legacy)",
		"Grain extract (legacy)",
		"Grain merge (legacy)",
		"Color erase (legacy)",
		"Overlay",
		"Hue (LCH)",
		"Chroma (LCH)",
		"Color (LCH)",
		"Lightness (LCH)",
		"Normal",
		"Behind",
		"Multiply",
		"Screen",
		"Difference",
		"Addition",
		"Substract",
		"Darken only",
		"Lighten only",
		"Hue (HSV)",
		"Saturation (HSV)",
		"Color (HSL)",
		"Value (HSV)",
		"Divide",
		"Dodge",
		"Burn",
		"Hard light",
		"Soft light",
		"Grain extract",
		"Grain merge",
		"Vivid light",
		"Pin light",
		"Linear light",
		"Hard mix",
		"Exclusion",
		"Linear burn",
		"Luma/Luminance darken only",
		"Luma/Luminance lighten only",
		"Luminance",
		"Color erase",
		"Erase",
		"Merge",
		"Split",
		"Pass through",
	]

	PROP_END: int = 0
	PROP_COLORMAP = 1
	PROP_ACTIVE_LAYER = 2
	PROP_ACTIVE_CHANNEL = 3
	PROP_SELECTION = 4
	PROP_FLOATING_SELECTION = 5
	PROP_OPACITY = 6
	PROP_MODE = 7
	PROP_VISIBLE = 8
	PROP_LINKED = 9
	PROP_LOCK_ALPHA = 10
	PROP_APPLY_MASK = 11
	PROP_EDIT_MASK = 12
	PROP_SHOW_MASK = 13
	PROP_SHOW_MASKED = 14
	PROP_OFFSETS = 15
	PROP_COLOR = 16
	PROP_COMPRESSION = 17
	PROP_GUIDES = 18
	PROP_RESOLUTION = 19
	PROP_TATTOO = 20
	PROP_PARASITES = 21
	PROP_UNIT = 22
	PROP_PATHS = 23
	PROP_USER_UNIT = 24
	PROP_VECTORS = 25
	PROP_TEXT_LAYER_FLAGS = 26
	PROP_OLD_SAMPLE_POINTS = 27
	PROP_LOCK_CONTENT = 28
	PROP_GROUP_ITEM = 29
	PROP_ITEM_PATH = 30
	PROP_GROUP_ITEM_FLAGS = 31
	PROP_LOCK_POSITION = 32
	PROP_FLOAT_OPACITY = 33
	PROP_COLOR_TAG = 34
	PROP_COMPOSITE_MODE = 35
	PROP_COMPOSITE_SPACE = 36
	PROP_BLEND_SPACE = 37
	PROP_FLOAT_COLOR = 38
	PROP_SAMPLE_POINTS = 39
	PROP_NUM_PROPS = 40

	def __init__(self, parent):
		"""A specialized binary file base for Gimp files."""
		self.parent = parent
		self.parasites: list[GimpParasite] = []
		self.guidelines: list[tuple[bool, int]] = []
		self.itemPath: list[str] | None = None
		self.vectors: list[GimpVector] = []
		self.colorMap: list[tuple[int, int, int]] = []
		self.userUnits: GimpUserUnits | None = None
		self.samplePoints: list[tuple[int, int]] = []
		self.selected: bool = False
		self.isSelection: bool = False
		self.selectionAttachedTo: str | None = None
		self.blendMode: int = 0  # one of self.BLEND_MODES
		self.visible: bool = False
		self.isLinked: bool = False
		self.lockAlpha: bool = False
		self.applyMask: bool = False
		self.editingMask: bool = False
		self.showMask: bool = False
		self.showMasked: bool = False
		self.xOffset: int = 0
		self.yOffset: int = 0
		self.compression: int = 0  # one of self.COMPRESSION_MODES
		self.horizontalResolution = None
		self.verticalResolution = None
		self.uniqueId = None
		self.units: int = 0  # one of self.UNITS
		self.textLayerFlags = None
		self.locked = None
		self.isGroup = None
		self.groupItemFlags: int = 0
		self.positionLocked: bool = False
		self.opacity: float = 1.0
		self.colorTag: int = 0  # one of self.TAG_COLORS
		self.compositeMode: int = 0  # one of self.COMPOSITE_MODES
		self.compositeSpace: int = 0  # one of self.COMPOSITE_SPACES
		self.blendSpace = None
		self.color = None
		self.vectorsVersion: int = 0
		self.activeVectorIndex: int = 0
		self.paths = []

	def getBlendMode(self) -> str:
		"""Return the blend mode as a string."""
		return self.BLEND_MODES[self.blendMode]

	def getCompression(self) -> str:
		"""Return the compression as a string."""
		return self.COMPRESSION_MODES[self.compression]

	def getUnits(self) -> str:
		"""Return the units as a string."""
		return self.UNITS[self.units]

	def getTagColours(self):
		"""Return the tag colours as a string."""
		return self.TAG_COLORS[self.colorTag]

	def getCompositeModes(self):
		"""Return the composite mode as a string."""
		return self.COMPOSITE_MODES[abs(self.compositeMode)]

	def getCompositeSpaces(self):
		"""Return the composite spaces as a string."""
		return self.COMPOSITE_SPACES[abs(self.compositeSpace)]

	@property
	def pointerSize(self) -> int:
		"""Determine the size of the "pointer" datatype based on the document version.

		NOTE: prior to version 11, it was 32-bit,
			since then it is 64-bit, thus supporting
			larger image files
		"""
		if self.doc.version >= 11:
			return 64
		return 32

	def _pointerDecode(self, ioBuf: IO) -> int:
		if self.pointerSize == 64:
			return ioBuf.u64
		return ioBuf.u32

	def _pointerEncode(self, ptr: int, ioBuf: IO | None = None) -> bytearray:
		if ioBuf is None:
			ioBuf = IO()
		if self.pointerSize == 64:
			ioBuf.u64 = ptr
		else:
			ioBuf.u32 = ptr
		return ioBuf.data

	@property
	def doc(self):
		"""Get the main document object."""
		item = self
		while item.parent != item:
			item = item.parent
		return item

	@property
	def root(self):
		"""Get the root of the file object tree (Which is the same as self.doc)."""
		return self.doc

	@property
	def tattoo(self):
		"""Gimp nomenclature for the item's unique id."""
		return self.uniqueId

	@tattoo.setter
	def tattoo(self, tattoo):
		"""Gimp nomenclature for the item's unique id."""
		self.uniqueId = tattoo

	def _parasitesDecode(self, data: bytes) -> int:
		"""Decode list of parasites."""
		index: int = 0
		while index < len(data):
			parasite = GimpParasite()
			index = parasite.decode(data, index)
			self.parasites.append(parasite)
		return index

	def _parasitesEncode(self):
		"""Encode list of parasites."""
		ioBuf = IO()
		for parasite in self.parasites:
			ioBuf.addBytes(parasite.encode())
		return ioBuf.data

	def _guidelinesDecode(self, data):
		"""Decode guidelines."""
		index: int = 0
		while index < len(data):
			position = struct.unpack(">I", data[index : index + 4])[0]
			index += 4
			isVertical = struct.unpack(">c", data[index])[0] == 2
			index += 1
			self.guidelines.append((isVertical, position))

	def _itemPathDecode(self, data):
		"""Decode item path."""
		index: int = 0
		path = []
		while index < len(data):
			pathElem = struct.unpack(">I", data[index : index + 4])[0]
			index += 4
			path.append(pathElem)
		self.itemPath = path

	def _vectorsDecode(self, data):
		"""Decode vectors."""
		index: int = 0
		self.vectorsVersion = struct.unpack(">I", data[index : index + 4])[0]
		index += 4
		self.activeVectorIndex = struct.unpack(">I", data[index : index + 4])[0]
		index += 4
		numPaths = struct.unpack(">I", data[index : index + 4])[0]
		index += 4
		for _ in range(numPaths):
			gimpV = GimpVector(self)
			gimpV.decode(data)
			self.vectors.append(gimpV)

	@property
	def activeVector(self):
		"""Get the vector that is currently active."""
		return self.vectors[self.activeVectorIndex]

	@property
	def expanded(self) -> bool:
		"""Is the group layer expanded."""
		return self.groupItemFlags & 0x00000001 > 0

	@expanded.setter
	def expanded(self, expanded):
		"""Is the group layer expanded."""
		if expanded:
			self.groupItemFlags |= 0x00000001
		else:
			self.groupItemFlags &= ~0x00000001

	def _colormapDecode(self, data, index=None):
		"""_colormapDecode_.

		:param data: can be bytes or an IO object

		decode colormap/palette
		"""
		ioObj = None
		if isinstance(data, IO):
			ioObj = data
			index = data.index
			data = data.data
		_ = struct.unpack(">I", data[0:4])[0]  # number of colors
		index += 4
		colors = []
		while index < len(data):
			red = data[index]
			index += 1
			green = data[index]
			index += 1
			blue = data[index]
			index += 1
			colors.append((red, green, blue))
		self.colorMap = colors
		if ioObj is not None:
			ioObj.index = index

	def _userUnitsDecode(self, data):
		"""Decode a set of user-defined measurement units."""
		userUnits = GimpUserUnits()
		userUnits.decode(data)
		self.userUnits = userUnits

	def _samplePointsDecode(self, data):
		"""Decode a series of points."""
		index: int = 0
		samplePoints = []
		while index < len(data):
			x = struct.unpack(">I", data[index : index + 4])[0]
			index += 4
			y = struct.unpack(">I", data[index : index + 4])[0]
			index += 4
			samplePoints.append((x, y))
		self.samplePoints = samplePoints

	def _propertyDecode(self, propertyType, data):
		"""Decode a single property.

		Many properties are in the form
		propertyType: one of PROP_
		lengthOfData: 4
		data: varies but is often ioBuf.32 or ioBuf.boolean
		"""
		ioBuf = IO(data, boolSize=32)
		# print('DECODING PROPERTY',propertyType,len(data))
		if propertyType == self.PROP_COLORMAP:
			self._colormapDecode(ioBuf)
		elif propertyType == self.PROP_ACTIVE_LAYER:
			self.selected = True
		elif propertyType == self.PROP_ACTIVE_CHANNEL:
			self.selected = True
		elif propertyType == self.PROP_SELECTION:
			self.isSelection = True
		elif propertyType == self.PROP_FLOATING_SELECTION:
			self.selectionAttachedTo = ioBuf.u32
		elif propertyType == self.PROP_OPACITY:
			self.opacity = ioBuf.u32
		elif propertyType == self.PROP_MODE:
			self.blendMode = ioBuf.u32
		elif propertyType == self.PROP_VISIBLE:
			self.visible = ioBuf.boolean
		elif propertyType == self.PROP_LINKED:
			self.isLinked = ioBuf.boolean
		elif propertyType == self.PROP_LOCK_ALPHA:
			self.lockAlpha = ioBuf.boolean
		elif propertyType == self.PROP_APPLY_MASK:
			self.applyMask = ioBuf.boolean
		elif propertyType == self.PROP_EDIT_MASK:
			self.editingMask = ioBuf.boolean
		elif propertyType == self.PROP_SHOW_MASK:
			self.showMask = ioBuf.boolean
		elif propertyType == self.PROP_SHOW_MASKED:
			self.showMasked = ioBuf.boolean
		elif propertyType == self.PROP_OFFSETS:
			self.xOffset = ioBuf.i32
			self.yOffset = ioBuf.i32
		elif propertyType == self.PROP_COLOR:
			red = ioBuf.byte
			green = ioBuf.byte
			blue = ioBuf.byte
			self.color = [red, green, blue]
		elif propertyType == self.PROP_COMPRESSION:
			self.compression = ioBuf.byte
		elif propertyType == self.PROP_GUIDES:
			self._guidelinesDecode(data)
		elif propertyType == self.PROP_RESOLUTION:
			self.horizontalResolution = ioBuf.float32
			self.verticalResolution = ioBuf.float32
		elif propertyType == self.PROP_TATTOO:
			self.uniqueId = data.hex()
		elif propertyType == self.PROP_PARASITES:
			self._parasitesDecode(data)
		elif propertyType == self.PROP_UNIT:
			self.units = ioBuf.u32
		elif propertyType == self.PROP_PATHS:
			_numPaths = ioBuf.u32
			"""
			for _ in range(numPaths):
				nRead, path = self._pathDecode_(data[index:])
				self.paths.append(path)
				index += nRead
			"""
		elif propertyType == self.PROP_USER_UNIT:
			self._userUnitsDecode(data)
		elif propertyType == self.PROP_VECTORS:
			pass
			# self._vectorsDecode_(data)
		elif propertyType == self.PROP_TEXT_LAYER_FLAGS:
			if isinstance(data, bytes):
				self.textLayerFlags = int.from_bytes(data, byteorder="big")
			else:
				self.textLayerFlags = int(data)
		elif propertyType == self.PROP_OLD_SAMPLE_POINTS:
			raise RuntimeError("ERR: old sample points structure not supported")
		elif propertyType == self.PROP_LOCK_CONTENT:
			self.locked = ioBuf.boolean
		elif propertyType == self.PROP_GROUP_ITEM:
			self.isGroup = True
		elif propertyType == self.PROP_ITEM_PATH:
			self._itemPathDecode(data)
		elif propertyType == self.PROP_GROUP_ITEM_FLAGS:
			self.groupItemFlags = ioBuf.u32
		elif propertyType == self.PROP_LOCK_POSITION:
			self.positionLocked = ioBuf.boolean
		elif propertyType == self.PROP_FLOAT_OPACITY:
			self.opacity = ioBuf.float32
		elif propertyType == self.PROP_COLOR_TAG:
			self.colorTag = ioBuf.u32
		elif propertyType == self.PROP_COMPOSITE_MODE:
			self.compositeMode = ioBuf.i32
		elif propertyType == self.PROP_COMPOSITE_SPACE:
			self.compositeSpace = ioBuf.i32
		elif propertyType == self.PROP_BLEND_SPACE:
			self.blendSpace = ioBuf.u32
		elif propertyType == self.PROP_FLOAT_COLOR:
			red = ioBuf.float32
			green = ioBuf.float32
			blue = ioBuf.float32
			self.color = [red, green, blue]
		elif propertyType == self.PROP_SAMPLE_POINTS:
			self._samplePointsDecode(data)
		else:
			raise RuntimeError(f"Unknown property id {propertyType}")
		return ioBuf.index

	def _propertyEncode(self, propertyType):
		"""Encode a single property.

		Many properties are in the form
		propertyType: one of PROP_
		lengthOfData: 4
		data: varies but is often ioBuf.32 or ioBuf.boolean

		If the property is the same as the default, or not specified, returns empty array
		"""
		ioBuf = IO(boolSize=32)
		if propertyType == self.PROP_COLORMAP:
			if self.colorMap is not None and self.colorMap:
				ioBuf.u32 = self.PROP_COLORMAP
				# ioBuf.addBytes(self._colormapEncode_())
		elif propertyType == self.PROP_ACTIVE_LAYER:
			if self.selected is not None and self.selected:
				ioBuf.u32 = self.PROP_ACTIVE_LAYER
		elif propertyType == self.PROP_ACTIVE_CHANNEL:
			if self.selected is not None and self.selected:
				ioBuf.u32 = self.PROP_ACTIVE_LAYER
		elif propertyType == self.PROP_SELECTION:
			if self.isSelection is not None and self.isSelection:
				ioBuf.u32 = self.PROP_SELECTION
		elif propertyType == self.PROP_FLOATING_SELECTION:
			if self.selectionAttachedTo is not None:
				ioBuf.u32 = self.PROP_FLOATING_SELECTION
				ioBuf.u32 = self.selectionAttachedTo
		elif propertyType == self.PROP_OPACITY:
			if self.opacity is not None and not isinstance(self.opacity, float):
				ioBuf.u32 = self.PROP_OPACITY
				ioBuf.u32 = self.opacity
		elif propertyType == self.PROP_MODE:
			if self.blendMode is not None:
				ioBuf.u32 = self.PROP_MODE
				ioBuf.u32 = self.blendMode
		elif propertyType == self.PROP_VISIBLE:
			if self.visible is not None:
				ioBuf.u32 = self.PROP_VISIBLE
				ioBuf.boolean = self.visible
		elif propertyType == self.PROP_LINKED:
			if self.isLinked is not None and self.isLinked:
				ioBuf.u32 = self.PROP_LINKED
				ioBuf.boolean = self.isLinked
		elif propertyType == self.PROP_LOCK_ALPHA:
			if self.lockAlpha is not None and self.lockAlpha:
				ioBuf.u32 = self.PROP_LOCK_ALPHA
				ioBuf.boolean = self.lockAlpha
		elif propertyType == self.PROP_APPLY_MASK:
			if self.applyMask is not None:
				ioBuf.u32 = self.PROP_APPLY_MASK
				ioBuf.boolean = self.applyMask
		elif propertyType == self.PROP_EDIT_MASK:
			if self.editingMask is not None and self.editingMask:
				ioBuf.u32 = self.PROP_EDIT_MASK
				ioBuf.boolean = self.editingMask
		elif propertyType == self.PROP_SHOW_MASK:
			if self.showMask is not None and self.showMask:
				ioBuf.u32 = self.PROP_SHOW_MASK
				ioBuf.boolean = self.showMask
		elif propertyType == self.PROP_SHOW_MASKED:
			if self.showMasked is not None:
				ioBuf.u32 = self.PROP_SHOW_MASKED
				ioBuf.boolean = self.showMasked
		elif propertyType == self.PROP_OFFSETS:
			if self.xOffset is not None and self.yOffset is not None:
				ioBuf.u32 = self.PROP_OFFSETS
				ioBuf.i32 = self.xOffset
				ioBuf.i32 = self.yOffset
		elif propertyType == self.PROP_COLOR:
			if (
				self.color is not None
				and not isinstance(self.color[0], float)
				and not isinstance(self.color[1], float)
				and not isinstance(self.color[2], float)
			):
				ioBuf.u32 = self.PROP_COLOR
				ioBuf.byte = self.color[0]
				ioBuf.byte = self.color[1]
				ioBuf.byte = self.color[2]
		elif propertyType == self.PROP_COMPRESSION:
			if self.compression is not None:
				ioBuf.u32 = self.PROP_COMPRESSION
				ioBuf.u32 = self.compression
		elif propertyType == self.PROP_GUIDES:
			if self.guidelines is not None and self.guidelines:
				pass
				# ioBuf.u32 = self.PROP_GUIDES
				# ioBuf.addBytes(self._guidelinesEncode_())
		elif propertyType == self.PROP_RESOLUTION:
			if self.horizontalResolution is not None and self.verticalResolution is not None:
				ioBuf.u32 = self.PROP_RESOLUTION
				ioBuf.u32 = int(self.horizontalResolution)
				ioBuf.i32 = int(self.verticalResolution)
		elif propertyType == self.PROP_TATTOO:
			if self.uniqueId is not None:
				ioBuf.u32 = int(self.uniqueId, 16)
		elif propertyType == self.PROP_PARASITES:
			if self.parasites is not None and self.parasites:
				ioBuf.u32 = self.PROP_PARASITES
				ioBuf.addBytes(self._parasitesEncode())
		elif propertyType == self.PROP_UNIT:
			if self.units is not None:
				ioBuf.u32 = self.PROP_UNIT
				ioBuf.u32 = self.units
		elif propertyType == self.PROP_PATHS:
			if self.paths is not None and self.paths:
				# ioBuf.u32 = self.PROP_PATHS
				# ioBuf.u32 = len(self.paths)
				"""
				for path in self.paths:
					ioBuf.append(self._pathEncode_(path))
				"""
		elif propertyType == self.PROP_USER_UNIT:
			if self.userUnits is not None:
				pass
				# ioBuf.u32 = self.PROP_USER_UNIT
				# ioBuf.addBytes(self._userUnitsEncode_())
		elif propertyType == self.PROP_VECTORS:
			if self.vectors is not None and self.vectors:
				pass
				# ioBuf.u32 = self.PROP_VECTORS
				# ioBuf.addBytes(self._vectorsEncode_())
		elif propertyType == self.PROP_TEXT_LAYER_FLAGS:
			if self.textLayerFlags is not None:
				ioBuf.u32 = self.PROP_TEXT_LAYER_FLAGS
				ioBuf.u32 = self.textLayerFlags
		elif propertyType == self.PROP_OLD_SAMPLE_POINTS:
			pass
		elif propertyType == self.PROP_LOCK_CONTENT:
			if self.locked is not None and self.locked:
				ioBuf.u32 = self.PROP_LOCK_CONTENT
				ioBuf.boolean = self.locked
		elif propertyType == self.PROP_GROUP_ITEM:
			if self.isGroup is not None and self.isGroup:
				ioBuf.u32 = self.PROP_GROUP_ITEM
		elif propertyType == self.PROP_ITEM_PATH:
			if self.itemPath is not None:
				pass
				# ioBuf.u32 = self.PROP_ITEM_PATH
				# ioBuf.addBytes(self._itemPathEncode_())
		elif propertyType == self.PROP_GROUP_ITEM_FLAGS:
			if self.groupItemFlags is not None:
				ioBuf.u32 = self.PROP_GROUP_ITEM_FLAGS
				ioBuf.u32 = self.groupItemFlags
		elif propertyType == self.PROP_LOCK_POSITION:
			if self.positionLocked is not None and self.positionLocked:
				ioBuf.u32 = self.PROP_LOCK_POSITION
				ioBuf.boolean = self.positionLocked
		elif propertyType == self.PROP_FLOAT_OPACITY:
			if self.opacity is not None and isinstance(self.opacity, float):
				ioBuf.u32 = self.PROP_FLOAT_OPACITY
				ioBuf.float32 = self.opacity
		elif propertyType == self.PROP_COLOR_TAG:
			if self.colorTag is not None:
				ioBuf.u32 = self.PROP_COLOR_TAG
				ioBuf.u32 = self.colorTag
		elif propertyType == self.PROP_COMPOSITE_MODE:
			if self.compositeMode is not None:
				ioBuf.u32 = self.PROP_COMPOSITE_MODE
				ioBuf.i32 = self.compositeMode
		elif propertyType == self.PROP_COMPOSITE_SPACE:
			if self.compositeSpace is not None:
				ioBuf.u32 = self.PROP_COMPOSITE_SPACE
				ioBuf.i32 = self.compositeSpace
		elif propertyType == self.PROP_BLEND_SPACE:
			if self.blendSpace is not None:
				ioBuf.u32 = self.PROP_BLEND_SPACE
				ioBuf.u32 = self.blendSpace
		elif propertyType == self.PROP_FLOAT_COLOR:
			if (
				self.color is not None
				and isinstance(self.color[0], float)
				and isinstance(self.color[1], float)
				and isinstance(self.color[2], float)
			):
				ioBuf.u32 = self.PROP_FLOAT_COLOR
				ioBuf.float32 = self.color[0]
				ioBuf.float32 = self.color[1]
				ioBuf.float32 = self.color[2]
		elif propertyType == self.PROP_SAMPLE_POINTS:
			if self.samplePoints is not None and self.samplePoints:
				pass
				# ioBuf.u32 = self.PROP_SAMPLE_POINTS
				# self.addBytes(self._samplePointsEncode_())
		else:
			raise RuntimeError(f"Unknown property id {propertyType}")
		return ioBuf.data

	def _propertiesDecode(self, ioBuf: IO):
		"""Decode a list of properties."""
		while True:
			try:
				propertyType = ioBuf.u32
				dataLength = ioBuf.u32
			except struct.error:  # end of data, so that's that.
				break
			if propertyType == 0:
				break
			self._propertyDecode(propertyType, ioBuf.getBytes(dataLength))
		return ioBuf.index

	def _propertiesEncode(self):
		"""Encode a list of properties."""
		ioBuf = IO()
		for propertyType in range(1, self.PROP_NUM_PROPS):
			moData = self._propertyEncode(propertyType)
			if moData:
				ioBuf.addBytes(moData)
		return ioBuf.data

	def __repr__(self, indent: str = ""):
		"""Get a textual representation of this object."""
		ret: list[str] = []
		attrs = [
			"itemPath",
			"selected",
			"isSelection",
			"selectionAttachedTo",
			"visible",
			"isLinked",
			"lockAlpha",
			"applyMask",
			"editingMask",
			"showMask",
			"showMasked",
			"uniqueId",
			"textLayerFlags",
			"locked",
			"isGroup",
			"groupItemFlags",
			"positionLocked",
			"opacity",
			"blendSpace",
		]
		for attr in attrs:
			_ = "" if getattr(self, attr) is None else ret.append(f"{attr}: {getattr(self, attr)}")

		if self.blendMode is not None:
			ret.append("Blend Mode: " + self.getBlendMode())

		if self.xOffset is not None:
			ret.append(f"Offset: {self.xOffset} x " + str(self.yOffset))
		if self.compression is not None:
			ret.append("Compression: " + self.getCompression())
		if self.horizontalResolution is not None:
			res = str(self.horizontalResolution) + f"ppi x {self.verticalResolution}ppi"
			ret.append(f"Resolution: {res}")
		if self.units is not None:
			ret.append("Units: " + self.getUnits())
		if self.colorTag is not None:
			ret.append("Tag Color: " + self.getTagColours())
		if self.compositeMode is not None:
			auto = "Auto " if self.compositeMode < 0 else ""  # negative values are "Auto"
			ret.append("Composite Mode: " + auto + self.getCompositeModes())
		if self.compositeSpace is not None:
			auto = "Auto " if self.compositeSpace < 0 else ""  # negative values are "Auto"
			ret.append("Composite Space: " + auto + self.getCompositeSpaces())
		if self.color is not None:
			ret.append(
				"Color: ("
				+ str(self.color[0])
				+ ","
				+ str(self.color[1])
				+ ","
				+ str(self.color[2])
				+ ")"
			)
		if self.userUnits is not None:
			ret.append("User Units: ")
			ret.append(self.userUnits.__repr__(indent + "\t"))
		if self.parasites:
			ret.append("Parasites: ")
			for item in self.parasites:
				ret.append(item.__repr__(indent + "\t"))
		if self.guidelines:
			ret.append("Guidelines: ")
			for item in self.guidelines:
				ret.append(item.__repr__())
		if self.samplePoints:
			ret.append("Sample Points: ")
			for item in self.samplePoints:
				ret.append(f"({item[0]}," + str(item[1]) + ")")
		if self.vectors:
			ret.append("Vectors: ")
			for item in self.vectors:
				ret.append(item.__repr__(indent + "\t"))
		if self.colorMap:
			ret.append("Color Map: ")
			for i, color in enumerate(self.colorMap):
				ret.append(str(i) + f": ({color[0]}," + str(color[1]) + f",{color[2]})")
		return indent + ((f"\n{indent}").join(ret))


class GimpUserUnits:
	"""User-defined measurement units."""

	def __init__(self):
		self.factor: int = 0
		self.numDigits: int = 0
		self.id = ""
		self.symbol = ""
		self.abbrev = ""
		self.sname = ""
		self.pname = ""

	def decode(self, data: bytes, index: int = 0) -> int:
		"""Decode a byte buffer.

		Args:
			data (bytes): data buffer to decode
			index (int, optional): index within the buffer to start at]. Defaults to 0.

		Returns:
			int: offset
		"""
		ioBuf = IO(data, index)
		self.factor = ioBuf.float32
		self.numDigits = ioBuf.u32
		self.id = ioBuf.sz754
		self.symbol = ioBuf.sz754
		self.abbrev = ioBuf.sz754
		self.sname = ioBuf.sz754
		self.pname = ioBuf.sz754
		return ioBuf.index

	def encode(self):
		"""Convert this object to raw bytes."""
		ioBuf = IO()
		ioBuf.float32 = self.factor
		ioBuf.u32 = self.numDigits
		ioBuf.sz754 = self.id
		ioBuf.sz754 = self.symbol
		ioBuf.sz754 = self.abbrev
		ioBuf.sz754 = self.sname
		ioBuf.sz754 = self.pname
		return ioBuf.data

	def __repr__(self, indent: str = ""):
		"""Get a textual representation of this object."""
		ret = []
		ret.append(f"Factor: {self.factor}")
		ret.append(f"Num Digits: {self.numDigits}")
		ret.append(f"ID: {self.id}")
		ret.append(f"Symbol: {self.symbol}")
		ret.append(f"Abbreviation: {self.abbrev}")
		ret.append(f"Singular Name: {self.sname}")
		ret.append(f"Plural Name: {self.pname}")
		return indent + ((f"\n{indent}").join(ret))



#!/usr/bin/env python3
"""Pure python implementation of the gimp gbr brush format.
"""
from __future__ import annotations

from io import BytesIO
from pathlib import Path

import PIL.Image
from binaryiotools import IO

from . import utils


class GimpGbrBrush:
	"""Pure python implementation of the gimp gbr brush format.

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/gbr.txt
	"""

	COLOR_MODES = [None, "L", "LA", "RGB", "RGBA"]  # only L or RGB allowed

	def __init__(self, fileName: str = None):
		"""Pure python implementation of the gimp gbr brush format.

		Args:
			fileName (str, optional): filename for the brush. Defaults to None.
		"""
		self.fileName = None
		self.version = 2
		self.width = 0
		self.height = 0
		self.bpp = 1
		self.mode = self.COLOR_MODES[self.bpp]
		self.name = ""
		self.rawImage = None
		self.spacing = 0
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: BytesIO | str):
		"""Load a gimp file.

		:param fileName: can be a file name or a file-like object
		"""
		self.fileName, data = utils.fileOpen(fileName)
		self.decode(data)

	def decode(self, data: bytes, index: int = 0) -> int:
		"""Decode a byte buffer.

		Args:
			data (bytes): data buffer to decode
			index (int, optional): index within the buffer to start at. Defaults to 0.

		Raises:
			RuntimeError: "unknown brush version"
			RuntimeError: "File format error.  Magic value mismatch"

		Returns:
			int: offset]
		"""
		ioBuf = IO(data, index)
		headerSize = ioBuf.u32
		self.version = ioBuf.u32
		if self.version != 2:
			raise RuntimeError(f"ERR: unknown brush version {self.version}")
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		self.bpp = ioBuf.u32  # only allows grayscale or RGB
		self.mode = self.COLOR_MODES[self.bpp]
		magic = ioBuf.getBytes(4)
		if magic.decode("ascii") != "GIMP":
			raise RuntimeError(
				'"'
				+ magic.decode("ascii")
				+ '" '
				+ str(index)
				+ " File format error.  Magic value mismatch."
			)
		self.spacing = ioBuf.u32
		nameLen = headerSize - ioBuf.index
		self.name = ioBuf.getBytes(nameLen).decode("UTF-8")
		self.rawImage = ioBuf.getBytes(self.width * self.height * self.bpp)
		return ioBuf.index

	def encode(self) -> bytearray:
		"""Encode this object to byte array."""
		ioBuf = IO()
		ioBuf.u32 = 28 + len(self.name)
		ioBuf.u32 = self.version
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		ioBuf.u32 = self.bpp
		ioBuf.addBytes("GIMP")
		ioBuf.u32 = self.spacing
		ioBuf.addBytes(self.name)
		ioBuf.addBytes(self.rawImage)
		return ioBuf.data

	@property
	def size(self) -> tuple[int, int]:
		"""Get the size."""
		return (self.width, self.height)

	@property
	def image(self) -> PIL.Image.Image | None:
		"""Get a final, compiled image."""
		if self.rawImage is None:
			return None
		return PIL.Image.frombytes(self.mode, self.size, self.rawImage, decoder_name="raw")

	def save(self, tofileName: str, toExtension: str | None = None):
		"""Save this gimp image to a file."""
		asImage = False
		if toExtension is None:
			if tofileName is not None:
				extParts = tofileName.rsplit(".", 1)
				if len(extParts) > 0:
					toExtension = extParts[-1]
				else:
					toExtension = None
		if toExtension is not None and toExtension != "gbr":
			asImage = True
		if asImage and self.image:
			self.image.save(tofileName)
			self.image.close()
		else:
			Path(tofileName).write_bytes(self.encode())

	def __repr__(self, indent=""):
		"""Get a textual representation of this object."""
		ret = []
		if self.fileName is not None:
			ret.append(f"fileName: {self.fileName}")
		ret.append(f"Name: {self.name}")
		ret.append(f"Version: {self.version}")
		ret.append(f"Size: {self.width} x {self.height}")
		ret.append(f"Spacing: {self.spacing}")
		ret.append(f"BPP: {self.bpp}")
		ret.append(f"Mode: {self.mode}")
		return (f"\n{indent}").join(ret)



#!/usr/bin/env python3
"""Pure python implementation of the OLD gimp gpb brush format.
"""
from __future__ import annotations

from io import BytesIO

from binaryiotools import IO

from . import utils
from .GimpGbrBrush import GimpGbrBrush
from .GimpPatPattern import GimpPatPattern


class GimpGpbBrush:
	"""Pure python implementation of the OLD gimp gpb brush format.

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/vbr.txt
	"""

	def __init__(self, fileName: BytesIO | str):
		"""Pure python implementation of the OLD gimp gpb brush format.

		Args:
			fileName (BytesIO): filename/ filepointer
		"""
		self.brush = GimpGbrBrush()
		self.pattern = GimpPatPattern()
		if isinstance(fileName, str):
			self.fileName = fileName
		else:
			self.fileName = fileName.name

	def load(self, fileName: BytesIO | str):
		"""Load a gimp file.

		:param fileName: can be a file name or a file-like object
		"""
		self.fileName, data = utils.fileOpen(fileName)
		self.decode(data)

	def decode(self, data: bytes, index: int = 0):
		"""Decode a byte buffer.

		Args:
			data (bytes): data to decode
			index (int, optional): index to start from. Defaults to 0.

		Returns:
			int: pointer
		"""
		index = self.brush.decode(data, index)
		# index = self.pattern.decode(data, index)
		return index

	def encode(self):
		"""Encode this object to a byte array."""
		ioBuf = IO()
		ioBuf.addBytes(self.brush.encode())
		ioBuf.addBytes(self.pattern.encode())
		return ioBuf.data

	def save(self, tofileName=None):
		"""Save this gimp image to a file."""
		utils.save(self.encode(), tofileName)

	def __repr__(self, indent=""):
		"""Get a textual representation of this object."""
		ret = []
		if self.fileName is not None:
			ret.append(f"fileName: {self.fileName}")
		ret.append(self.brush.__repr__(indent + "\t"))
		ret.append(self.pattern.__repr__())
		return (f"\n{indent}").join(ret)



from __future__ import annotations

from io import BytesIO


def fileOpen(fileName: BytesIO | str) -> tuple[str, bytes]:
	if isinstance(fileName, str):
		file = open(fileName, "rb")
	else:
		fileName = fileName.name
		file = fileName
	data = file.read()
	file.close()
	return fileName, data


def save(data: bytes, tofileName: BytesIO | str):
	"""Save this gimp image to a file."""
	if isinstance(tofileName, str):
		file = open(tofileName, "wb")
	else:
		file = tofileName
	file.write(data)
	file.close()



#!/usr/bin/env python3
"""Pure python implementation of the gimp vbr brush format.
"""
from __future__ import annotations

from io import BytesIO

from . import utils


class GimpVbrBrush:
	"""Pure python implementation of the gimp vbr brush format.

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/vbr.txt
	"""

	BRUSH_SHAPES = ["circle", "square", "diamond"]

	def __init__(self, fileName: BytesIO | str | None = None):
		"""Pure python implementation of the gimp vbr brush format.

		Args:
			fileName (BytesIO, str, optional): filename. Defaults to None.
		"""
		self.version = 1.0
		self.name = ""
		self.spacing = 0
		self.radius = 50
		self.hardness = 1
		self.aspectRatio = 1
		self.angle = 0
		self.brushShape = None  # one of the strings in self.BRUSH_SHAPES
		self.spikes = None
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: BytesIO | str):
		"""Load a gimp file.

		:param fileName: can be a file name or a file-like object
		"""
		self.fileName, data = utils.fileOpen(fileName)
		self.decode(data)

	@property
	def image(self):
		"""This parametric brush converted to a useable PIL image."""
		raise NotImplementedError()  # TODO:

	def decode(self, dataIn: bytes):
		"""Decode a byte buffer.

		:param dataIn: data buffer to decode
		"""
		data = [s.strip() for s in dataIn.decode("utf-8").split("\n")]
		if data[0] != "GIMP-VBR":
			raise RuntimeError("File format error.  Magic value mismatch.")
		self.version = float(data[1])
		if self.version == 1.0:
			self.name = data[2]  # max len 255 bytes
			self.spacing = float(data[3])
			self.radius = float(data[4])
			self.hardness = float(data[5])
			self.aspectRatio = float(data[6])
			self.angle = float(data[7])
		elif self.version == 1.5:
			self.name = data[2]  # max len 255 bytes
			self.brushShape = data[3]  # one of the strings in self.BRUSH_SHAPES
			self.spacing = float(data[4])
			self.radius = float(data[5])
			self.spikes = float(data[6])
			self.hardness = float(data[7])
			self.aspectRatio = float(data[8])
			self.angle = float(data[9])
		else:
			raise RuntimeError(f"Unknown version {self.version}")

	def encode(self):
		"""Encode to a raw data stream."""
		data = []
		data.append("GIMP-VBR")
		data.append(str(self.version))
		if self.version == 1.0:
			data.append(str(self.name))
			data.append(str(self.spacing))
			data.append(str(self.radius))
			data.append(str(self.hardness))
			data.append(str(self.aspectRatio))
			data.append(str(self.angle))
		elif self.version == 1.5:
			data.append(str(self.name))
			data.append(str(self.brushShape))
			data.append(str(self.spacing))
			data.append(str(self.radius))
			data.append(str(self.spikes))
			data.append(str(self.hardness))
			data.append(str(self.aspectRatio))
			data.append(str(self.angle))
		return ("\n".join(data) + "\n").encode("utf-8")

	def save(self, tofileName=None, toExtension=None):
		"""Save this gimp image to a file."""
		asImage = False
		if toExtension is None:
			if tofileName is not None:
				toExtension = tofileName.rsplit(".", 1)
				if len(toExtension) > 1:
					toExtension = toExtension[-1]
				else:
					toExtension = None
		if toExtension is not None and toExtension != "vbr":
			asImage = True
		if asImage:
			self.image.save(tofileName)
		else:
			if hasattr(tofileName, "write"):
				file = tofileName
			else:
				file = open(tofileName, "wb")
			file.write(self.encode())
			file.close()

	def __repr__(self):
		"""Get a textual representation of this object."""
		ret = []
		if self.fileName is not None:
			ret.append(f"fileName: {self.fileName}")
		ret.append(f"Name: {self.name}")
		ret.append(f"Version: {self.version}")
		ret.append(f"Spacing: {self.spacing}")
		ret.append(f"Radius: {self.radius}")
		ret.append(f"Hardness: {self.hardness}")
		ret.append(f"Aspect ratio: {self.aspectRatio}")
		ret.append(f"Angle: {self.angle}")
		ret.append(f"Brush Shape: {self.brushShape}")
		ret.append(f"Spikes: {self.spikes}")
		return "\n".join(ret)

	def __eq__(self, other):
		"""Perform a comparison."""
		return all(
			(
				other.name == self.name,
				other.version == self.version,
				other.spacing == self.spacing,
				other.radius == self.radius,
				other.hardness == self.hardness,
				other.aspectRatio == self.aspectRatio,
				other.angle == self.angle,
				other.brushShape == self.brushShape,
				other.spikes == self.spikes,
			)
		)



"""Gets packed pixels from a gimp image.

NOTE: This was originally designed to be a hierarchy, like
	an image pyramid, through in practice they only use the
	top level of the pyramid (64x64) and ignore the rest.
"""
from __future__ import annotations

from PIL import Image

from .GimpImageLevel import GimpImageLevel
from .GimpIOBase import IO, GimpIOBase


class GimpImageHierarchy(GimpIOBase):
	"""
	Gets packed pixels from a gimp image

	NOTE: This was originally designed to be a hierarchy, like
		an image pyramid, through in practice they only use the
		top level of the pyramid (64x64) and ignore the rest.
	"""

	def __init__(self, parent, image: Image.Image | None = None):
		GimpIOBase.__init__(self, parent)
		self.width: int = 0
		self.height: int = 0
		self.bpp: int = 0  # Number of bytes per pixel given
		self._levelPtrs = []
		self._levels = None
		self._data = None
		if image is not None:  # NOTE: can override earlier parameters
			self.image = image

	def decode(self, data: bytes, index: int = 0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		if not data:
			raise RuntimeError("No data!")
		ioBuf = IO(data, index)
		# print 'Decoding channel at',index
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		self.bpp = ioBuf.u32
		if self.bpp < 1 or self.bpp > 4:
			msg = (
				"""'Unexpected bytes-per-pixel for image data ("""
				+ str(self.bpp)
				+ """).
				Probably means file corruption."""
			)
			raise RuntimeError(msg)
		while True:
			ptr = self._pointerDecode(ioBuf)
			if ptr == 0:
				break
			self._levelPtrs.append(ptr)
		if self._levelPtrs:  # remove "dummy" level pointers
			self._levelPtrs = [self._levelPtrs[0]]
		self._data = data
		return ioBuf.index

	def encode(self):
		"""Encode this object to a byte buffer."""
		dataioBuf = IO()
		ioBuf = IO()
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		ioBuf.u32 = self.bpp
		dataIndex = ioBuf.index + self.pointerSize * (len(self.levels) + 1)
		for level in self.levels:
			ioBuf.addBytes(
				self._pointerEncode(dataIndex + dataioBuf.index)
			)  # TODO: This may be incorrect
			dataioBuf.addBytes(level.encode())
		ioBuf.addBytes(self._pointerEncode(0))
		ioBuf.addBytes(dataioBuf.data)
		return ioBuf.data

	@property
	def levels(self):
		"""Get the levels within this hierarchy.

		Presently hierarchy is not really used by gimp,
		so this returns an array of one item
		"""
		if self._levels is None:
			for ptr in self._levelPtrs:
				level = GimpImageLevel(self)
				level.decode(self._data, ptr)
				self._levels = [level]
		return self._levels

	@property
	def image(self) -> Image.Image | None:
		"""Get a final, compiled image."""
		if not self.levels:
			return None
		return self.levels[0].image

	@image.setter
	def image(self, image: Image.Image):
		"""Set the image."""
		self.width = image.width
		self.height = image.height
		if image.mode not in ["L", "LA", "RGB", "RGBA"]:
			raise NotImplementedError("Unsupported PIL image type")
		self.bpp = len(image.mode)
		self._levelPtrs = None
		# self._levels = [GimpImageLevel(self, image)]

	def __repr__(self, indent: str = ""):
		"""Get a textual representation of this object."""
		ret = []
		ret.append(f"Size: {self.width} x {self.height}")
		ret.append(f"Bytes Per Pixel: {self.bpp}")
		return indent + ((f"\n{indent}").join(ret))



#!/usr/bin/env python3
"""Gimp Image Pipe Format.

The gih format is use to store a series of brushes, and some extra info
for how to use them.
"""
from __future__ import annotations

from io import BytesIO

from binaryiotools import IO

from . import utils
from .GimpGbrBrush import GimpGbrBrush


class GimpGihBrushSet:
	"""Gimp Image Pipe Format.

	The gih format is use to store a series of brushes, and some extra info
	for how to use them.

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/gih.txt
	"""

	def __init__(self, fileName: str = None):
		"""Gimp Image Pipe Format.

		Args:
			fileName (str, optional): filename. Defaults to None.
		"""
		self.fileName = None
		self.name = ""
		self.params = {}
		self.brushes = []
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: BytesIO | str):
		"""Load a gimp file.

		:param fileName: can be a file name or a file-like object
		"""
		self.fileName, data = utils.fileOpen(fileName)
		self.decode(data)

	def decode(self, data: bytes, index: int = 0) -> int:
		"""Decode a byte buffer.

		Args:
			data (bytes): data buffer to decode
			index (int, optional): index within the buffer to start at. Defaults to 0.

		Returns:
			int: offset
		"""
		ioBuf = IO(data, index)
		self.name = ioBuf.textLine
		secondLine = ioBuf.textLine.split(" ")
		self.params = {}
		numBrushes = int(secondLine[0])
		# everything that's left is a gimp-image-pipe-parameters parasite
		for i in range(1, len(secondLine)):
			param = secondLine[i].split(":", 1)
			self.params[param[0].strip()] = param[1].strip()
		self.brushes = []
		for _ in range(numBrushes):
			brush = GimpGbrBrush()
			ioBuf.index = brush.decode(
				ioBuf.data, ioBuf.index
			)  # TODO: broken.  For some reason there is extra data between brushes!
			self.brushes.append(brush)
		return ioBuf.index

	def encode(self):
		"""Encode this object to a byte array."""
		ioBuf = IO()
		ioBuf.textLine = self.name
		# add the second line of data
		secondLine = [str(len(self.brushes))]
		for key, value in self.params.items():
			secondLine.append(f"{key}:{value}")
		ioBuf.textLine = " ".join(secondLine)
		# add the brushes
		for brush in self.brushes:
			ioBuf.addBytes(brush.encode())
		return ioBuf.data

	def save(self, tofileName: str):
		"""Save this gimp image to a file."""
		utils.save(self.encode(), tofileName)

	def __repr__(self, indent=""):
		"""Get a textual representation of this object."""
		ret = []
		if self.fileName is not None:
			ret.append(f"fileName: {self.fileName}")
		ret.append(f"Name: {self.name}")
		for k, val in list(self.params.items()):
			ret.append(k + f": {val}")
		for i, brush in enumerate(self.brushes):
			ret.append(f"Brush {i}")
			ret.append(brush.__repr__(indent + "\t"))
		return (f"\n{indent}").join(ret)
